import numpy as np
import pandas as pd

import utils as ut


class FractionCollector:
    """
    A high-level wrapper around an XY stage.
    """
    def __init__(self, xy):
        self.XY = xy

        # TODO: should transforms be a property of the position table?
        self.frames = pd.DataFrame(index=['transform', 'position_table'])
        self.add_frame('hardware')
        # TODO: load in deck transform
        deck_transform = np.array([[-1, 0, 0],
                                   [0, 1, 0],
                                   [0, 0, 0]])
        self.add_frame('deck', deck_transform)
        # TODO: option to add plates here

    def add_frame(self, name, transform=np.eye(3, 2), position_table=None):
        # self.frames.assign(name=[np.eye(4,3),None]) # returns new frame with appended col
        self.frames[name] = None
        self.frames[name].transform = transform
        self.frames[name].position_table = position_table

    # ref_frame gives reference frame; its offset is ignored
    def add_plate(self, name, filepath, ref_frame='deck'):
        # move to bottom of first well, i.e. plate origin (using GUI)

        # store offset (translation)
        offset = self.where()  # hardware_xy - plate_xy (0,0,0)

        # determine transform
        transform = self.frames[ref_frame].copy()
        transform[-1] = offset

        # add position_table - either:
        # A) add from file
        while True:
            filepath = raw_input('Enter filepath to plate position_table:')
            try:
                position_table = ut.read_delim_pd(filepath)
                break
            except IOError:
                print 'No file:', filepath

        # add frame
        self.add_frame(name, transform, position_table)

    def where(self):
        """
        Retrieves current (X,Y) position.

        :return: (tup) current position
        """
        return self.XY.where_xy()

    def home(self):
        """
        Homes XY axes.
        """
        self.XY.home_xy()

    # TODO: if no columns specified, transform provided XY to hardware coordinates.
    # TODO: default frame?
    def goto(self, frame, lookup_columns, lookup_values):
        """
        Finds lookup_values in lookup_columns of frame's position_list; retrieves corresponding X,Y
        Transforms X,Y to hardware X,Y by frame's transform.
        Moves to hardware X,Y

        :param frame: (str) frame that specifies position_list and transform
        :param lookup_columns: (str | list) column(s) to search in position_table
        :param lookup_values: (val | list) values(s) to find in lookup_columns
        """
        transform, position_table = self.frames[frame]

        xy = tuple(ut.lookup(position_table, lookup_columns, lookup_values)[['x', 'y']].iloc[0])
        xyw = xy + (1,)  # concatenate for translation
        xh, yh = np.dot(xyw, transform)  # get hardware coordinates

        self.XY.goto_xy(xh, yh)

    def exit(self):
        """
        Send exit command to XY.
        """
        self.XY.exit()



import time

import serial as s
import yaml


class AsiController:
    """
    Low-level wrapper for the Applied Scientific Instrumentation (ASI) Controller.
    Config file must be defined.

    This class can control both an XY-axis (MS-2000 stage) and a Z-axis (LS-50 linear stage).
    Since this hardware was taken from an Illumina GaIIx, it assumes the controller's serial command-set requires an OEM
    prefix of 1h (Z-axis) or 2h (XY-axes). Both stages have a linear-encoder.

    Functions for Z-axis control are defined, but it is not initialized. If it is desired to be used, then a homing
    procedure needs to be defined in initialize().
    """
    def __init__(self, config_file, init=True):
        self.serial = s.Serial()  # placeholder
        
        f = open(config_file, 'r')
        self.config = yaml.load(f)
        f.close()
        
        self.config['conv'] = float(self.config['conv'])

        if init:
            self.initialize()

    def initialize(self):
        """
        1) Open serial connection
        2) Enable XY and Z motor control (one-time command)
        3) Move to XY stage stops (hall-effect)
        4) Move from switch limits 0.2 mm
        5) Zero stage
        """
        self.serial = s.Serial(**self.config['serial'])  # open serial connection

        self.cmd_xy('MC x+ y+')  # enable motor control for xy
        self.cmd_z('MC z+')      # enable motor control for z

        print "Initializing stage..."
        # TODO: should homing / initialization this be responsibility of the autosipper?
        # TODO: Implement zeroing commands?
        self.goto_xy(1000, -1000)  # move to switch limits (bottom right)
        self.move_relative_xy(-0.2, 0.2)  # move from switch limits 0.2 mm
        self.cmd_xy('HERE x y')  # establish current XY position as 0,0 (zero)

    def cmd(self, cmd_string):
        """
        Wraps core cmd_string with terminator specified in config, writes to serial, and returns response.

        :param cmd_string: (str) core command (w/o prefix nor terminator)
        :return: (str) device response
        """
        full_string = self.config['prefix'] + cmd_string + self.config['terminator']
        self.serial.write(full_string)
        time.sleep(0.05)
        response = self.serial.read(self.serial.inWaiting())
        return response
    
    def halt(self):
        """
        Sends halt command to both axes, interrupting execution of their current commands.
        Note that many commands are sent in 'blocking' mode, so this function will likely not be called until the
        axes finish executing their current command.

        In the future, it may be nice to implement a 'waiting' scheme.
        """
        self.halt_xy()
        self.halt_z()

    def exit(self):
        """
        Closes the device's serial connection.
        """
        self.serial.close()
    
    # XY ----------------------------------------------
    def cmd_xy(self, cmd_string, block=True):
        """
        Wraps core cmd_string with axes prefix (2h), passes to the cmd() function, and returns response.
        Optionally blocks programmatic flow (default=True).

        :param cmd_string: (str) core command (w/o prefix nor terminator)
        :param block: (bool) whether the command blocks program flow until action is complete
        :return: (str) device response
        """
        full_string = '2h ' + cmd_string
        response = self.cmd(full_string)
        
        while block and self.is_busy_xy():
            time.sleep(0.05)
            pass
         
        return response

    def is_busy_xy(self):
        """
        Sends status command, then parses response to determine if XY-axes are busy.

        :return: (bool) true if axes are executing a command
        """
        status = self.cmd('2h STATUS')[0]
        return status == 'B'

    def halt_xy(self):
        """
        Sends halt command to the XY-axes (stage), interrupting execution of its current command.
        Note that many commands are sent in 'blocking' mode, so this function will likely not be called until the
        axes finish executing their current command.

        In the future, it may be nice to implement a 'waiting' scheme.
        """
        self.cmd_xy('HALT', False)

    def home_xy(self):
        """
        Moves XY-axes to 0,0.
        """
        return self.goto_xy(0, 0)
    
    def where_xy(self):
        """
        Retrieves XY-axes' current position relative to zero point (w/ linear encoder).

        :return: (tup) current X and Y position [mm]
        """
        conv = self.config['conv']
        response = self.cmd_xy('WHERE X Y')
        if response.find('A'):
            pos_xy = response.split()[1:3]
            pos_x = round(float(pos_xy[0])/conv, 4)
            pos_y = round(float(pos_xy[1])/conv, 4)
            return pos_x, pos_y
        else:
            return None, None

    def goto_xy(self, x_mm, y_mm):
        """
        Moves XY-axes absolutely to the specified position.

        :param x_mm: (float) desired absolute X position [mm]
        :param y_mm: (float) desired absolute Y position [mm]
        :return: (str) device response
        """
        conv = self.config['conv']
        x_str = 'x=' + str(float(x_mm) * conv)
        y_str = 'y=' + str(float(y_mm) * conv)
        return self.cmd_xy(' '.join(['m', x_str, y_str]))
    
    def move_relative_xy(self, x_mm, y_mm):
        """
        Moves XY-axes relatively by the specified number of mm.

        :param x_mm: (float) desired relative movement [mm]
        :param y_mm: (float) desired relative movement [mm]
        :return: (str) device response
        """
        conv = self.config['conv']
        x_str = 'x=' + str(float(x_mm) * conv)
        y_str = 'y=' + str(float(y_mm) * conv)
        return self.cmd_xy(' '.join(['r', x_str, y_str]))

    # Z -----------------------------------------------
    def cmd_z(self, cmd_string, block=True):
        """
        Wraps core cmd_string with axis prefix (1h), passes to the cmd() function, and returns response.
        Optionally blocks programmatic flow (default=True).

        :param cmd_string: (str) core command (w/o prefix nor terminator)
        :param block: (bool) whether the command blocks program flow until action is complete
        :return: (str) device response
        """
        while block and self.is_busy_z():
            time.sleep(0.3)
        full_string = '1h ' + cmd_string
        return self.cmd(full_string)
    
    def is_busy_z(self):
        """
        Sends status command, then parses response to determine if Z-axis is busy.

        :return: (bool) true if axis is executing a command
        """
        status = self.cmd('1h STATUS')
        return status[0] == 'B'

    def halt_z(self):
        """
        Sends halt command to the Z-axis (linear motor), interrupting execution of its current command.
        Note that many commands are sent in 'blocking' mode, so this function will likely not be called until the
        axes finish executing their current command.

        In the future, it may be nice to implement a 'waiting' scheme.
        """
        self.cmd_z('HALT', False)

    def home_z(self):
        """
        Moves Z-axis to 0.
        """
        return self.goto_z(0)
        
    def where_z(self):
        """
        Retrieves Z-axis' current position relative to zero point (w/ linear encoder).

        :return: (tup) current Z position [mm]
        """
        response = self.cmd_z('WHERE Z')
        if response.find('A'):
            pos_z = float(response.split()[1:2])
            return pos_z
        else:
            return None    
    
    def goto_z(self, z_mm):
        """
        Moves Z-axis absolutely to the specified position.

        :param z_mm: (float) desired absolute Z position [mm]
        :return: (str) device response
        """
        conv = self.config['conv']
        z_str = 'z=' + str(float(z_mm) * conv)
        return self.cmd_z(' '.join(['m', z_str]))
    
    def move_relative_z(self, z_mm):
        """
        Moves Z-axis relatively by the specified number of mm.

        :param z_mm: (float) desired relative movement [mm]
        :return: (str) device response
        """
        conv = self.config['conv']
        z_str = 'z=' + str(float(z_mm) * conv)
        return self.cmd_z(' '.join(['r', z_str]))



import csv
import json

import numpy as np
import pandas as pd


def read_delim(filepath):
    """
    Reads delimited file (auto-detects delimiter + header). Returns list.

    :param filepath: (str) location of delimited file
    :return: (list) list of records w/o header
    """
    f = open(filepath, 'r')
    dialect = csv.Sniffer().sniff(f.read(1024))
    f.seek(0)
    has_header = csv.Sniffer().has_header(f.read(1024))
    f.seek(0)
    reader = csv.reader(f, dialect)

    if has_header:
        reader.next()

    ret = [line for line in reader]
    return ret


def read_delim_pd(filepath):
    """
    Reads delimited file (auto-detects delimiter + header). Returns pandas DataFrame.

    :param filepath: (str) location of delimited file
    :return: (DataFrame)
    """
    f = open(filepath)
    has_header = None
    if csv.Sniffer().has_header(f.read(1024)):
        has_header = 0
    f.seek(0)
    return pd.read_csv(f, header=has_header, sep=None, engine='python')


def lookup(table, lookup_cols, lookup_vals, output_cols=None, output_recs=None):
    """
    Looks up records where lookup_cols == lookup_vals.
    Optionally returns only specified output_cols and/or specified output_recs.

    :param table: (DataFrame) the pandas DataFrame to use as a lookup table
    :param lookup_cols: (str | list)
    :param lookup_vals: (val | list)
    :param output_cols:
    :param output_recs:
    :return:
    """
    if type(lookup_cols) == str:
        lookup_cols = [lookup_cols]

    lookup_vals = [lookup_vals]
    temp_df = pd.DataFrame(data=lookup_vals, columns=lookup_cols, copy=False)
    output = table.merge(temp_df, copy=False)

    if output_cols is not None:
        if type(output_cols) == str:
            output_cols = [output_cols]
        output = output[output_cols]

    if output_recs is not None:
        output = output.iloc[output_recs]

    return output


def generate_position_table(num_rc, space_rc, z, to_clipboard=False):
    """
    Generates a position table.

    :param num_rc: (tup) number of rows and columns (num_rows, num_cols)
    :param space_rc: (tup) spacing for rows and columns [mm] (spacing_rows, spacing_cols)
    :param z: (float) z value in table [mm]
    :param to_clipboard: (bool) whether to copy the position_table to the OS clipboard
    :return: (DataFrame)
    """
    temp = list()
    headers = ['n', 's', 'r', 'c', 'name', 'x', 'y', 'z']

    for r in range(num_rc[0]):
        for c in range(num_rc[1]):
            n = c + r * num_rc[1]
            s = ((r + 1) % 2) * (c + r * num_rc[1]) + (r % 2) * ((r + 1) * num_rc[1] - (c + 1))
            name = chr(64 + r + 1) + '{:02d}'.format(c + 1)
            x = float(c * space_rc[1])
            y = float(r * space_rc[0])
            z = float(z)
            temp.append([n, s, r, c, name, x, y, z])
    position_table = pd.DataFrame(temp, columns=headers)
    if to_clipboard:
        position_table.to_clipboard(index=False)
    return position_table


def spacing(num_rc, p1, p2):
    r, c = map(float, num_rc)
    return tuple(abs(np.nan_to_num(np.subtract(p2, p1) / (c - 1, r - 1))))


def load_mm_positionlist(filepath):
    """
    Takes a MicroManager position list and converts it to a pandas DataFrame.

    :param filepath: (str)
    :return: (DataFrame) position list with headers = "r, c, name, x, y"
    """
    with open(filepath) as f:
        data = json.load(f)

    df1 = pd.io.json.json_normalize(data, ['POSITIONS']).drop(
        ['DEFAULT_XY_STAGE', 'DEFAULT_Z_STAGE', 'DEVICES', 'PROPERTIES'], 1)
    df1 = df1[['GRID_ROW', 'GRID_COL', 'LABEL']]
    df2 = pd.io.json.json_normalize(data, ['POSITIONS', 'DEVICES']).drop(['DEVICE', 'AXES', 'Z'], 1)
    df = pd.concat([df1, df2], axis=1)

    rename = {'GRID_ROW': 'r',
              'GRID_COL': 'c',
              'LABEL': 'name',
              'X': 'x',
              'Y': 'y'}

    return df.rename(columns=rename)


def generate_grid(c0, c1, l_img, p):
    """
    Based on two points, creates a 2D-acquisition grid similar to what MicroManager would produce.

    :param c0: (arr) first point; numpy 1d array of len 2
    :param c1: (arr) second point; numpy 1d array of len 2
    :param l_img: (float)
    :param p: (float) desired percent overlap
    :return: (DataFrame) position_list in the same format as load_mm_positionlist
    """
    # TODO: does generate_grid subsume generate_position_table?
    # n -> number of stage positions on an axis
    n = 1 + np.ceil(np.abs(c1 - c0) / ((1 - p) * l_img))  # ct,ct
    n = n.astype('int')

    # l_acq = total_movement + l_img
    # l_acq = l_img * (n - n*p + p)  # um,um
    sign = np.sign(c1 - c0)

    # could also use cartesian product (itertools.product OR np.mgrid, stack)
    # https://stackoverflow.com/questions/1208118/using-numpy-to-build-an-array-of-all-combinations-of-two-arrays
    position_list = pd.DataFrame(columns=['r', 'c', 'name', 'x', 'y'], )
    for j in xrange(n[1]):  # iter y
        y = sign[1] * j * l_img * (1 - p) + c0[1]
        for i in xrange(n[0]) if not (j % 2) else reversed(xrange(n[0])):  # iter x (serp)
            x = sign[0] * i * l_img * (1 - p) + c0[0]

            r = j
            c = i
            name = '1-Pos_{:03}_{:03}'.format(c, r)
            position_list.loc[len(position_list)] = [r, c, name, x, y]

    position_list[['r', 'c']] = position_list[['r', 'c']].astype(int)
    return position_list



import time

import serial as s
import yaml


# TODO: get current position for relative move
class Motor:
    """
    Low-level wrapper for the Lin Engineering (LE) CO-4118S-09.
    Config file must be defined.

    The LE CO-4118S-09 has an integrated controller with a documented serial command-set. It lacks an encoder, and so
    relies on dead-reckoning for position. It does have an optical sensor that allows it to get a positional fix (home).
    """
    def __init__(self, config_file, init=True):
        self.serial = s.Serial()  # placeholder

        f = open(config_file, 'r')
        self.config = yaml.load(f)
        f.close()

        self.config['conv'] = float(self.config['conv'])

        if init:
            self.initialize()

    def initialize(self):
        """
        1) Open serial connection
        2) Set velocity to velocity specified in config file
        3) Home the motor
        """
        self.serial = s.Serial(**self.config['serial'])  # open serial connection

        # TODO set moving current
        # TODO set holding current
        self.set_velocity(self.config['velocity_limit'])  # set velocity
        self.home()  # move motor to home

    def cmd(self, cmd_string, block=True):
        """
        Wraps core cmd_string with prefix and terminator specified in config, writes to serial, and returns response.
        Optionally blocks programmatic flow (default=True).

        :param cmd_string: (str) core command (w/o prefix nor terminator)
        :param block: (bool) whether the command blocks program flow until action is complete
        :return: (str) device response
        """
        full_string = self.config['prefix'] + cmd_string + self.config['terminator']
        self.serial.write(full_string)

        time.sleep(0.15)  # TODO: monitor for response?
        response = self.serial.read(self.serial.inWaiting()).decode('utf8', 'ignore')

        while block and self.is_busy():
            pass

        return response

    def is_busy(self):
        """
        Sends query command, then parses response to determine if motor is busy.

        :return: (bool) true if motor is executing a command
        """
        cmd_string = 'Q'
        time.sleep(0.05)
        response = self.cmd(cmd_string, False)
        return response.rfind('`') == -1

    def set_velocity(self, velocity):
        """
        Checks requested velocity against the velocity limit, then sets motor velocity in usteps/sec.

        :param velocity: (int) velocity
        :return: (str) device response
        """
        if velocity > self.config['velocity_limit']:
            velocity = self.config['velocity_limit']
            print 'ERR: Desired velocity exceeds velocity_limit; velocity now set to velocity_limit'

        cmd_string = 'V{}R'.format(velocity)
        return self.cmd(cmd_string)

    def halt(self):
        """
        Sends halt command to motor, which stops it from executing its current command.
        Note that many commands are sent in 'blocking' mode, so this function will likely not be called until the
        motor finishes executing its current command.

        In the future, it may be nice to implement a 'waiting' scheme.
        """
        cmd_string = 'T'
        self.cmd(cmd_string)

    def home(self):
        """
        Homes the motor until the optical sensor is triggered. Zero position is reset (motor gets positional fix).

        :return: (str) device response
        """
        cmd_string = 'Z{}R'.format(self.config['ustep_max'])
        return self.cmd(cmd_string)

    def goto(self, mm, block=True):
        """
        Moves motor absolutely to the specified position.

        :param mm: (float) desired absolute position [mm]
        :param block: (bool) whether the command blocks program flow until action is complete
        :return: (str) device response
        """
        ustep = int(self.config['conv'] * mm)

        if ustep > self.config['ustep_max']:
            ustep = self.config['ustep_max']
            print 'ERR: Desired move to {} mm exceeds max of {} mm; moving to max instead'.format(mm, self.config[
                'ustep_max'] / self.config['conv'])
        if ustep < self.config['ustep_min']:
            ustep = self.config['ustep_min']
            print 'ERR: Desired move to {} mm exceeds min of {} mm; moving to min instead'.format(mm, self.config[
                'ustep_min'] / self.config['conv'])

        cmd_string = 'A{}R'.format(ustep)

        return self.cmd(cmd_string, block)

    def move_relative(self, mm):
        """
        Moves motor relatively by the specified number of mm.

        :param mm: (float) desired relative movement [mm]
        :return: (str) device response
        """
        ustep = int(self.config['conv'] * mm)
        ustep_current = int(self.config['ustep_max'] / 2)  # TODO: limit movement (+ and -)

        if mm >= 0:
            if (ustep_current + ustep) > self.config['ustep_max']:
                ustep = self.config['ustep_max'] - ustep_current
                print 'ERR: Desired move of +{} mm exceeds max of {} mm; moving to max instead'.format(mm, self.config[
                    'ustep_max'] / self.config['conv'])
            cmd_string = 'P{}R'.format(ustep)

        else:
            if (ustep_current + ustep) < self.config['ustep_min']:
                ustep = self.config['ustep_min'] - ustep_current
                print 'ERR: Desired move of {} mm exceeds min of {} mm; moving to min instead'.format(mm, self.config[
                    'ustep_min'] / self.config['conv'])
            ustep = -1 * ustep
            cmd_string = 'D{}R'.format(ustep)

        return self.cmd(cmd_string)

    def where(self):
        """
        Retrieves motor's current position relative to zero-position (by dead reckoning).

        :return: (tup) current position of the motor [mm]
        """
        cmd_string = '?0'
        response = str(self.cmd(cmd_string))
        strt = response.rfind('`') + 1
        for end, c in enumerate(response[strt:]):
            if c not in str(range(0, 11)):
                break
        ustep = response[strt:strt + end]
        return round(float(ustep) / self.config['conv'], 4),  # tuple

    def exit(self):
        """
        Closes the device's serial connection.
        """
        self.serial.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

install_requirements = [
    # TODO: put package requirements here
    'numpy',
    'pandas',
    'matplotlib',
    'ipywidgets',
    'opencv-python',
    'opencv-contrib-python',
    'pymodbus',
    'pyserial',
    'PyYAML',
]

setup_requirements = [
    'pytest-runner',
    # TODO(slongwell): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    # TODO: put package test requirements here
]

setup(
    name='acqpack',
    version='0.8.0',
    description="Library for instrument control and automated data acquisition",
    long_description=readme + '\n\n' + history,
    author="Scott Longwell",
    author_email='longwell@stanford.edu',
    url='https://github.com/fordycelab/acqpack',
    packages=find_packages(include=['acqpack']),
    include_package_data=True,
    install_requires=install_requirements,
    license="MIT license",
    zip_safe=False,
    keywords='acqpack',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        # 'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



# -*- coding: utf-8 -*-

"""Top-level package for acqpack."""

__author__ = """Scott Longwell"""
__email__ = 'longwell@stanford.edu'
__version__ = '0.8.0'

import gui
import utils
from asicontroller import AsiController
from autosampler import Autosampler
from fractioncollector import FractionCollector
from manifold import Manifold
from motor import Motor

# TODO: packaging https://python-packaging.readthedocs.io/en/latest/non-code-files.html



import time

import cv2
import matplotlib.pyplot as plt
import numpy as np
from IPython import display
from ipywidgets import widgets


# --------------------------------------------------------------------  
# IMSHOW -------------------------------------------------------------
def imshow(img, name='Image', mode='cv2'):
    if mode == 'cv2':
        cv2.namedWindow(name, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(name, 500, 500)
        cv2.setWindowProperty(name, cv2.WND_PROP_ASPECT_RATIO, cv2.WINDOW_KEEPRATIO)
        cv2.imshow(name, img)
        k = cv2.waitKey(30)
    else:
        plt.imshow(img, cmap='gray')
        plt.show()


# --------------------------------------------------------------------  
# SNAP ---------------------------------------------------------------
def snap(core, mode='mpl'):
    def on_snap_mpl(b):
        core.snapImage()
        img = core.getImage()
        plt.imshow(img, cmap='gray')
        plt.show()

    def on_snap_cv2(b):
        cv2.destroyWindow('Snap')
        cv2.namedWindow('Snap', cv2.WINDOW_NORMAL)
        cv2.resizeWindow('Snap', 500, 500)
        cv2.setWindowProperty('Snap', cv2.WND_PROP_ASPECT_RATIO, cv2.WINDOW_KEEPRATIO)
        core.snapImage()
        time.sleep(.1)
        img = core.getImage()
        cv2.imshow('Snap', img)
        k = cv2.waitKey(30)

    def on_close_cv2(b):
        cv2.destroyWindow('Snap')

    snap = widgets.Button(description='Snap')

    if mode == 'cv2':
        snap.on_click(on_snap_cv2)
        close = widgets.Button(description='Close')
        close.on_click(on_close_cv2)
        display.display(widgets.HBox([snap, close]))
    else:  # mode=='mpl'
        snap.on_click(on_snap_mpl)
        display.display(snap)


# --------------------------------------------------------------------
# VIDEO ------------------------------------------------------
def video(core, loop_pause=0.15):
    # video with button (CV2)
    live = widgets.Button(description='Live')
    close = widgets.Button(description='Close')
    display.display(widgets.HBox([live, close]))

    def on_live(b):
        display.clear_output(wait=True)
        print 'LIVE'
        core.startContinuousSequenceAcquisition(1000)  # time overridden by exposure
        time.sleep(.2)
        cv2.namedWindow('Video', cv2.WINDOW_NORMAL)
        cv2.setWindowProperty('Video', cv2.WND_PROP_ASPECT_RATIO, cv2.WINDOW_KEEPRATIO)
        cv2.resizeWindow('Video', 500, 500)
        img = np.zeros((500, 500))
        print 'To stop, click window + press ESC'
        while 1:
            time.sleep(loop_pause)
            if core.getRemainingImageCount() > 0:
                img = core.getLastImage()
            cv2.imshow('Video', img)
            k = cv2.waitKey(30)
            if k == 27:  # ESC key; may need 255 mask?
                break
        print 'STOPPED'
        core.stopSequenceAcquisition()

    def on_close(b):
        if core.isSequenceRunning():
            core.stopSequenceAcquisition()
        cv2.destroyWindow('Video')

    live.on_click(on_live)
    close.on_click(on_close)

    # video with button (CV2)
    # serial snap image

    # live = widgets.Button(description='Live')
    # close = widgets.Button(description='Close')
    # display.display(widgets.HBox([live, close]))

    # def on_live_clicked(b):
    #     display.clear_output(wait=True)
    #     print 'LIVE'
    #     cv2.namedWindow('Video', cv2.WINDOW_NORMAL)
    #     cv2.setWindowProperty('Video', cv2.WND_PROP_ASPECT_RATIO, cv2.WINDOW_KEEPRATIO)
    #     cv2.resizeWindow('Video', 500,500)
    #     img = np.zeros((500,500))
    #     print 'To stop, click window + press ESC'
    #     while(1):
    #         core.snapImage()
    #         time.sleep(.05)
    #         img = core.getImage()
    #         cv2.imshow('Video',img)
    #         k = cv2.waitKey(30)
    #         if k==27: # ESC key; may need 255 mask?
    #             break
    #     print 'STOPPED'

    # def on_close_clicked(b):
    #     if core.isSequenceRunning():
    #         core.stopSequenceAcquisition()
    #     cv2.destroyWindow('Video')

    # live.on_click(on_live_clicked)
    # close.on_click(on_close_clicked)


# --------------------------------------------------------------------  
# GRID ------------------------------------------------------
def grid(core, loop_pause=0.15):
    # video


    # video with button (CV2)

    def on_live(b):
        display.clear_output(wait=True)
        print 'LIVE'
        core.startContinuousSequenceAcquisition(1000)  # time overridden by exposure
        time.sleep(.2)
        cv2.namedWindow('Video', cv2.WINDOW_NORMAL)
        cv2.setWindowProperty('Video', cv2.WND_PROP_ASPECT_RATIO, cv2.WINDOW_KEEPRATIO)
        cv2.resizeWindow('Video', 500, 500)
        img = np.zeros((500, 500))
        print 'To stop, click window + press ESC'
        while 1:
            time.sleep(loop_pause)
            if core.getRemainingImageCount() > 0:
                img = core.getLastImage()
            cv2.imshow('Video', img)
            k = cv2.waitKey(30)
            if k == 27:  # ESC key; may need 255 mask?
                break
        print 'STOPPED'
        core.stopSequenceAcquisition()

    def on_close(b):
        if core.isSequenceRunning():
            core.stopSequenceAcquisition()
        cv2.destroyWindow('Video')

    def on_corner(b):
        pos = core.get_xy()
        b.owner.value = pos
        print pos

    live = widgets.Button(description='Live')
    close = widgets.Button(description='Close')

    c1 = widgets.Button(description='Corner 1')
    c2 = widgets.Button(description='Corner 2')
    file = widgets.Text()
    save = widgets.Text()

    live.on_click(on_live)
    close.on_click(on_close)

    c1.on_click(on_corner)
    c2.on_click(on_corner)

    # calculate: c1.value - c2.value
    save.on_click(icon='floppy-o')

    display.display(widgets.HBox([live, close]))


# --------------------------------------------------------------------
# MANIFOLD -----------------------------------------------------------
def manifold_control(manifold, button_col='name'):
    def on_clicked(b):
        if b.new == True:  # keep this (b.new == True)
            manifold.depressurize(b.owner.valve)
        elif b.new == False:  # keep this (b.new == False)
            manifold.pressurize(b.owner.valve)
        else:
        	pass

    def sync(b):
        for button in button_list:
            button.value = manifold.read_valve(button.valve)

    button_list = []
    for i in range(48):
        desc = '{} . . . {}'.format(i, manifold.valvemap[button_col][i])
        button_list.append(
            widgets.ToggleButton(
                valve=i,
                description=desc,
                value=manifold.read_valve(i),
                button_style=''  # 'success', 'info', 'warning', 'danger' or ''
            ))

    for button in button_list:
        button.observe(on_clicked)

    bank_list = []
    for i in range(0, 48, 8):
        bank_list.append(
            widgets.VBox(button_list[i:i + 8][::-1])
        )

    sync_button = widgets.Button(icon='fa-retweet', button_style='success', layout=widgets.Layout(width='40px'))
    sync_button.on_click(sync)
    display.display(widgets.HBox(bank_list + [sync_button]))
    display.display(widgets.Label('Dark = De-Pressurized / Open', layout=widgets.Layout(width='300px')))

    
# --------------------------------------------------------------------
# STAGE CONTROL-------------------------------------------------------
def stage_control(autosampler):
    xy = autosampler.XY
    z = autosampler.Z
    # icons are from "font-awesome"

    # xy ------------------------------------------------------------
    x_minus = widgets.Button(
        ax='x',
        sign=-1,
        button_style='primary',
        icon='fa-arrow-left',
        width='50px')

    x_plus = widgets.Button(
        ax='x',
        sign=1,
        button_style='primary',
        icon='fa-arrow-right',
        width='50px')

    y_minus = widgets.Button(
        ax='y',
        sign=-1,
        button_style='primary',
        icon='fa-arrow-up',
        width='50px')

    y_plus = widgets.Button(
        ax='y',
        sign=1,
        button_style='primary',
        icon='fa-arrow-down',
        width='50px')

    xy_home = widgets.Button(
        ax='xy',
        button_style='primary',
        icon='fa-home',
        width='50px')

    xy_slider = widgets.FloatSlider(description='[mm]', min=.05, max=10, step=.05, orientation='vertical',
                                    height='150px')
    # xy_slider.observe(step, names='value')

    xy_cluster = widgets.HBox(
        [xy_slider, widgets.VBox([widgets.HBox([x_minus, x_plus, xy_home]), widgets.HBox([y_minus, y_plus])])])

    # z ------------------------------------------------------------
    z_minus = widgets.Button(
        ax='z',
        sign=-1,
        button_style='primary',
        icon='fa-arrow-up')

    z_plus = widgets.Button(
        ax='z',
        sign=1,
        button_style='primary',
        icon='fa-arrow-down')

    z_home = widgets.Button(
        ax='z',
        button_style='primary',
        icon='fa-home',
        width='50px')

    z_slider = widgets.FloatSlider(description='[mm]',
                                   min=.05, max=10,
                                   step=.05,
                                   orientation='vertical',
                                   height='150px')
    # z_slider.observe(zstep, names='value')

    z_cluster = widgets.VBox([widgets.HBox([z_slider, widgets.VBox([z_minus, z_plus]), z_home])])

    # type to move ------------------------------------------------------------
    x_pos = widgets.Text(
        ax='x',
        value='0',
        placeholder='enter pos',
        description='X:',
        width='150px')

    y_pos = widgets.Text(
        ax='y',
        value='0',
        placeholder='enter pos',
        description='Y:',
        width='150px')

    z_pos = widgets.Text(
        ax='z',
        value='0',
        placeholder='enter pos',
        description='Z:',
        width='150px')

    # functions ------------------------------------------------------------
    def move(b):
        if b.ax == 'x':
            xy.move_relative_xy(b.sign * xy_slider.value, 0)
        elif b.ax == 'y':
            xy.move_relative_xy(0, b.sign * xy_slider.value)
        elif b.ax == 'z':
            z.move_relative(b.sign * z_slider.value)

        display.clear_output()
        print(xy.where_xy() + z.where())

    for button in [x_minus, x_plus, y_minus, y_plus, z_minus, z_plus]:
        button.on_click(move)

    def home(b):
        if b.ax == 'xy':
            z.home()
            xy.home_xy()
        elif b.ax == 'z':
            z.home()

        display.clear_output()
        print(xy.where_xy() + z.where())

    for button in [xy_home, z_home]:
        button.on_click(home)

    def pos(b):
        if b.ax == 'x':
            x_curr, y_curr = xy.where_xy()
            xy.goto_xy(b.value, y_curr)
        if b.ax == 'y':
            x_curr, y_curr = xy.where_xy()
            xy.goto_xy(x_curr, b.value)
        elif b.ax == 'z':
            z.goto(float(b.value))

        display.clear_output()
        print(xy.where_xy() + z.where())

    for button in [x_pos, y_pos, z_pos]:
        button.on_submit(pos)
        
    # def step(change):
    #     xy_step = change['new']

    # for slider in [xy_slider, z_slider]:
    #     slider.observe(step, names='value')

    line = widgets.Label(value="$---------------------------------------$")
    print(xy.where_xy() + z.where())

    return widgets.VBox([widgets.HBox([x_pos, y_pos, z_pos]), line, widgets.HBox([xy_cluster, z_cluster])])



from pymodbus.client.sync import ModbusTcpClient

import utils as ut


class Manifold:
    """
    Provides a wrapper for the manifold, which is controlled by the Wago nModbus
    """
    def __init__(self, ip_address, valvemap_path, read_offset=512):
        self.client = ModbusTcpClient(ip_address)
        self.read_offset = read_offset
        self.valvemap = None
        self.load_valvemap(valvemap_path)

    def load_valvemap(self, valvemap_path):
        """
        Stores valvemap.
        To work with open/close, valvemap should have one column named 'valve'.

        :param valvemap_path: (str) path to valvemap
        """
        self.valvemap = ut.read_delim_pd(valvemap_path)

    def read_valve(self, valve_num):
        """
        Reads the state of the register associated with the specified valve.

        :param valve_num: (int) register number to read
        :return: () state of the register (True: depressurized, False: pressurized)
        """
        register_num = valve_num + self.read_offset
        return self.client.read_coils(register_num, 1).bits[0]

    def pressurize(self, valve_num):
        """
        Pressurizes valve at the specified register.

        :param valve_num: (int) valve to pressurize
        """
        state = self.read_valve(valve_num)
        if state:
            self.client.write_coil(valve_num, False)

    def depressurize(self, valve_num):
        """
        Depressurizes valve at the specified register.

        :param valve_num: (int) valve to depressurize
        """
        state = self.read_valve(valve_num)
        if not state:
            self.client.write_coil(valve_num, True)

    def close(self, lookup_cols, lookup_vals):
        """
        Finds lookup_vals in lookup_cols of valvemap; retrieves corresponding valve_num.
        Closes valve_num.

        :param lookup_cols: (str | list) column(s) to search in valvemap
        :param lookup_vals: (val | list) value(s) to find in lookup_cols
        """
        # TODO: default lookup_cols
        valve_num = ut.lookup(self.valvemap, lookup_cols, lookup_vals)[['valve']].iloc[0]
        self.pressurize(valve_num)

    def open(self, lookup_cols, lookup_vals):
        """
        Finds lookup_vals in lookup_cols of valvemap; retrieves corresponding valve_num.
        Opens valve_num.

        :param lookup_cols: (str | list) column(s) to search in valvemap
        :param lookup_vals: (val | list) value(s) to find in lookup_cols
        """
        valve_num = ut.lookup(self.valvemap, lookup_cols, lookup_vals)[['valve']].iloc[0]
        self.depressurize(valve_num)

    def exit(self):
        """
        Closes the device's serial connection.
        """
        self.client.close()
